/*
 * Original code, including ROM code, developed by Andrew Gregory.
 * MoDapple wrapper code and Windows adaptation by Steve Nickolas.
 */
#include <stdint.h>
#include <time.h>

static uint8_t clockrom[256]=
{
 0x08, 0x78, 0x28, 0x2C, 0x58, 0xFF, 0x70, 0x05, 0x18, 0x90, 0x7D, 0xD0, 0x65, 0x18, 0xB0, 0x38,
 0x48, 0x8A, 0x48, 0x98, 0x48, 0x08, 0x78, 0x20, 0x58, 0xFF, 0xBA, 0x68, 0x68, 0x68, 0x68, 0xA8,
 0xCA, 0x9A, 0x68, 0x28, 0xAA, 0x98, 0xB0, 0x14, 0xE4, 0x39, 0xD0, 0x04, 0xA4, 0x38, 0xF0, 0x03,
 0x18, 0x90, 0x40, 0xA9, 0x00, 0x9D, 0x38, 0x05, 0xA9, 0x0F, 0x85, 0x38, 0x8A, 0x0A, 0x0A, 0x0A,
 0x0A, 0xA8, 0xBD, 0x38, 0x05, 0x99, 0x8A, 0xC0, 0xB9, 0x8B, 0xC0, 0x48, 0xBD, 0x38, 0x05, 0xD9,
 0x89, 0xC0, 0xF0, 0x05, 0xFE, 0x38, 0x05, 0xD0, 0x09, 0xA9, 0x00, 0x9D, 0x38, 0x05, 0x68, 0xA9,
 0x8D, 0x48, 0x68, 0xA8, 0x08, 0x78, 0xBA, 0x68, 0x68, 0x68, 0x68, 0x98, 0x48, 0x9A, 0x28, 0x18,
 0x90, 0x10, 0x38, 0x08, 0x48, 0x8A, 0x0A, 0x0A, 0x0A, 0x0A, 0xA8, 0x68, 0x99, 0x88, 0xC0, 0x28,
 0xB0, 0x05, 0x68, 0xA8, 0x68, 0xAA, 0x68, 0x60, 0x48, 0x8A, 0x0A, 0x0A, 0x0A, 0x0A, 0xA8, 0x68,
 0x99, 0x88, 0xC0, 0xA9, 0x00, 0x9D, 0x38, 0x05, 0xA8, 0x8A, 0x48, 0xA9, 0xA8, 0x48, 0x48, 0x8A,
 0x48, 0x98, 0x48, 0x8A, 0x48, 0xA9, 0x3B, 0x48, 0x60, 0x99, 0x00, 0x02, 0xC9, 0x8D, 0xF0, 0x03,
 0xC8, 0xD0, 0xE6, 0xA9, 0x00, 0x99, 0x00, 0x02, 0x60, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
};

/*
 * ClockData[ 0x00 ] = Year
 * ClockData[ 0x01 ] = Century
 * ClockData[ 0x02 ] = Month
 * ClockData[ 0x03 ] = Day
 * ClockData[ 0x04 ] = Day of Week
 * ClockData[ 0x05 ] = Hour
 * ClockData[ 0x06 ] = Minute
 * ClockData[ 0x07 ] = Second
 *
 * ClockData[ 0x08 ] = Mode
 * ClockData[ 0x09 ] = String Length
 * ClockData[ 0x0a ] = Character Position
 * ClockData[ 0x0b ] = Character
 *
 */

static uint8_t ClockData[16];

/* Given a printf-like format string, build a string in buf that contains
   the appropriately formatted string. Format commands are as follows:
     %y = year
     %c = century
     %o = month
     %d = day
     %w = day of week
     %h = hour
     %m = minute
     %s = second
*/
static void FormatDateTime( char *buf, char *fmt )
{
    int idx, done;

    done = 0;
    while ( !done )
    {
        if ( *fmt == '%' )
        {
            /* Handle % formatting */
            fmt++; /* skip % */
            idx = -1;
            switch ( *fmt )
            {
            case 'Y':
            case 'y':
                idx = 0;
                break;
            case 'C':
            case 'c':
                idx = 1;
                break;
            case 'O':
            case 'o':
                idx = 2;
                break;
            case 'D':
            case 'd':
                idx = 3;
                break;
            case 'W':
            case 'w':
                idx = 4;
                break;
            case 'H':
            case 'h':
                idx = 5;
                break;
            case 'M':
            case 'm':
                idx = 6;
                break;
            case 'S':
            case 's':
                idx = 7;
                break;
            case 0:
                done = 1;
                break;
            }
            if ( idx!=-1 )
            {
                /* found a valid format specifier */
                *buf++ = ( ClockData[ idx ] >>   4 ) + '0';
                *buf++ = ( ClockData[ idx ] &  0xf ) + '0';
            }
            else
            {
                /* invalid format specifier: copy the character */
                *buf++ = *fmt;
            }
            fmt++;
        }
        else {
            /* not a % format */
            if ( *fmt == 0 )
            {
                done = 1;
            }
            *buf++ = *fmt++;
        }
    }
    *buf=0;
}

static uint8_t ReadClockIO( uint16_t Address )
{
    char string[ 256 ];
    int Loop;

    time_t curtime;
    struct tm *tm;
    
    time(&curtime);
    tm=localtime(&curtime);
    
    ClockData[0]=((tm->tm_year)+1900)&0xFF;
    ClockData[1]=(((tm->tm_year)+1900)>>8)&0xFF;
    ClockData[2]=tm->tm_mon+1;
    ClockData[3]=tm->tm_mday;
    ClockData[4]=tm->tm_wday;
    ClockData[5]=tm->tm_hour;
    ClockData[6]=tm->tm_min;
    ClockData[7]=tm->tm_sec;
    
    /* Now that we've got everything - convert it into BCD */
    Loop = ClockData[ 0x01 ] * 256 + ClockData[ 0x00 ];
    ClockData[ 0x00 ] = Loop /   10 % 10 * 16 + Loop       % 10;
    ClockData[ 0x01 ] = Loop / 1000 % 10 * 16 + Loop / 100 % 10;
    for ( Loop = 2; Loop < 8; Loop++ )
    {
        ClockData[ Loop ] = ClockData[ Loop ] / 10 % 10 * 16 +
                            ClockData[ Loop ]      % 10;
    }

    /* Build appropriately formatted string */
//  ; $A3 = '#' : ProDOS format "oo,ww,dd,hh,mm"
//  ; $BE = '>' : Integer BASIC format   "oo/dd  hh;mm;ss;.000" ?
//  ; $A0 = ' ' : Applesoft BASIC format "oo/dd  hh;mm;ss;.000"
//  ; $A5 = '%' : Applesoft BASIC format "oo/dd  hh;mm;ss;.000"
//  ; $A4 = '$' : My custom ProDOS clock format "yy/oo/dd hh:mm:ss" for use with
    switch ( ClockData[ 0x08 ] )
    {
    case 0xa3: /* ProDOS format */
        FormatDateTime( string, "%o,%w,%d,%h,%m" );
        break;
    case 0xbe:
    case 0xa0:
    case 0xa5: /* Applesoft BASIC format */
        FormatDateTime( string, " %o/%d  %h;%m;%s;.000" );
        break;
    case 0xa4: /* Custom ProDOS format */
        FormatDateTime( string, "%y/%o/%d %h;%m;%s" );
        break;
    }
    /* set high bit for all characters */
    Loop = 0;
    while ( string[ Loop ] != 0 )
    {
        string[ Loop++ ] |= 0x80;
    }
    /* set return character */
    ClockData[ 0x0b ] = string[ ClockData[ 0x0a ] ];

    /* Return appropriate value */
    return ClockData[ Address & 0x0f ];
}

#pragma argsused
static void WriteClockIO( uint16_t Address, uint8_t Data )
{
    switch ( Address & 0x0f )
    {
     case 0x08:
        /* Set mode */
        switch ( Data )
        {
        case 0xa3:
        case 0xbe:
        case 0xa0:
        case 0xa5:
        case 0xa4:
            ClockData[ 0x08 ] = Data;
            break;
        }
        /* Set lengths */
        switch ( ClockData[ 0x08 ] )
        {
        case 0xa3:
            ClockData[ 0x09 ] = 14;
            break;
        case 0xbe:
        case 0xa0:
        case 0xa5:
            ClockData[ 0x09 ] = 21;
            break;
        case 0xa4:
            ClockData[ 0x09 ] = 17;
            break;
        }
        break;
     case 0x0a:
        /* Set current character position */
        if ( Data < ClockData[ 0x09 ] )
        {
            ClockData[ 0x0a ] = Data;
        }
        break;
    }
    /* Update current clock data */
    ReadClockIO( Address );
}

__declspec(dllexport) uint32_t mod_type=0x544F4C53;
__declspec(dllexport) uint8_t *mod_name="Real Time Clock";
__declspec(dllexport) uint8_t *mod_pref="rtc";
__declspec(dllexport) uint8_t *mod_vers="0.01";
__declspec(dllexport) uint8_t *mod_copr="Copyright 2017 Steve Nickolas\nOriginal code by Andrew Gregory";

__declspec(dllexport) uint8_t  *rtc_preload_name=0;
__declspec(dllexport) uint16_t rtc_preload_size=0;
__declspec(dllexport) uint8_t  *rtc_preload_ptr=0;

__declspec(dllexport) void rtc_write16 (uint16_t a, uint8_t v)
{
 WriteClockIO(a,v);
}

__declspec(dllexport) void rtc_write256 (uint16_t a, uint8_t v)
{
}

__declspec(dllexport) void rtc_write2048 (uint16_t a, uint8_t v)
{
}

__declspec(dllexport) uint8_t rtc_read16 (uint16_t a)
{
 return ReadClockIO(a);
}

__declspec(dllexport) uint8_t rtc_read256 (uint16_t a)
{
 return clockrom[a&0xFF];
}

__declspec(dllexport) uint8_t rtc_read2048 (uint16_t a)
{
 return 255;
}

__declspec(dllexport) int16_t rtc_override_wr (uint16_t a, uint8_t v)
{
 return -1;
}

__declspec(dllexport) int16_t rtc_override_rd (uint16_t a)
{
 return -1;
}

__declspec(dllexport) int rtc_reset (void)
{
 return 0;
}

__declspec(dllexport) int rtc_loop (uint32_t cycles)
{
 return 0;
}

__declspec(dllexport) int rtc_init (uint8_t slot)
{
 return 0;
}

__declspec(dllexport) int rtc_deinit (void)
{
 return 0;
}

__declspec(dllexport) void rtc_config (void)
{
}
